import datetime
import time
import polygon
# from polygon import RESTClient
from sqlalchemy import create_engine
from sqlalchemy import text
import pandas as pd
from math import sqrt
from math import isnan
from numpy import mean
from numpy import std
from math import floor
import sys
from forex_data.return_classes import AUDUSD_return
from forex_data.return_classes import GBPEUR_return
from forex_data.return_classes import USDCAD_return
from forex_data.return_classes import USDJPY_return
from forex_data.return_classes import USDMXN_return
from forex_data.return_classes import EURUSD_return
from forex_data.return_classes import USDCNY_return
from forex_data.return_classes import USDCZK_return
from forex_data.return_classes import USDPLN_return
from forex_data.return_classes import USDINR_return


# Function slightly modified from polygon sample code to format the date string
def ts_to_datetime(ts) -> str:
	return datetime.datetime.fromtimestamp(ts / 1000.0).strftime('%Y-%m-%d %H:%M:%S')


# Function which clears the raw data tables once we have aggregated the data in a 6 minute interval
def reset_raw_data_tables(engine, currency_pairs):
	with engine.begin() as conn:
		for curr in currency_pairs:
			conn.execute(text("DROP TABLE " + curr[0] + curr[1] + "_raw;"))
			conn.execute(
				text("CREATE TABLE " + curr[0] + curr[1] + "_raw(ticktime text, fxrate  numeric, inserttime text);"))


# This creates a table for storing the raw, unaggregated price data for each currency pair in the SQLite database
def initialize_raw_data_tables(engine, currency_pairs):
	with engine.begin() as conn:
		for curr in currency_pairs:
			try:
				conn.execute(text(
					"CREATE TABLE " + curr[0] + curr[1] + "_raw(ticktime text, fxrate  numeric, inserttime text);"))
				print("executed create raw")
			except Exception as e:
				print(e)


# This creates a table for storing the (6 min interval) aggregated price data for each currency pair in the SQLite database
def initialize_aggregated_tables(engine, currency_pairs):
	with engine.begin() as conn:
		for curr in currency_pairs:
			try:
				conn.execute(text("CREATE TABLE " + curr[0] + curr[
					1] + "_agg(inserttime text, avgfxrate  numeric, stdfxrate numeric);"))
				print("executed create aggregated")
			except Exception as e:
				print(e)


def initialize_aggfive_tables(engine, curr):
	with engine.begin() as conn:
		# for curr in currency_pairs:
		try:
			conn.execute(text("CREATE TABLE " + curr[0] + curr[
				1] + "_aggfive(max numeric, min numeric, mean numeric, vol numeric, fd numeric, inserttime text);"))
			print("executed create aggregated")
		except Exception as e:
			print(e)


# This function is called every 6 minutes to aggregate the data, store it in the aggregate table,
# and then delete the raw data
def aggregate_raw_data_tables(engine, currency_pairs):
	with engine.begin() as conn:
		for curr in currency_pairs:
			print("aggregating curr: \n", curr)
			result = conn.execute(
				text("SELECT AVG(fxrate) as avg_price, COUNT(fxrate) as tot_count FROM " + curr[0] + curr[1] + "_raw;"))
			print("SELECT AVG(fxrate) as avg_price, COUNT(fxrate) as tot_count FROM " + curr[0] + curr[1] + "_raw;")

			#             print("avg_result: ", result)
			for row in result:
				#                 print(row)
				avg_price = row.avg_price
				tot_count = row.tot_count
			std_res = conn.execute(text(
				"SELECT SUM((fxrate - " + str(avg_price) + ")*(fxrate - " + str(avg_price) + "))/(" + str(
					tot_count) + "-1) as std_price FROM " + curr[0] + curr[1] + "_raw;"))
			#             print("std_res: ", std_res)
			for row in std_res:
				std_price = sqrt(row.std_price)
			date_res = conn.execute(text("SELECT MAX(ticktime) as last_date FROM " + curr[0] + curr[1] + "_raw;"))
			for row in date_res:
				last_date = row.last_date
			conn.execute(text("INSERT INTO " + curr[0] + curr[
				1] + "_agg (inserttime, avgfxrate, stdfxrate) VALUES (:inserttime, :avgfxrate, :stdfxrate);"),
			             [{"inserttime": last_date, "avgfxrate": avg_price, "stdfxrate": std_price}])

			# This calculates and stores the return values
			exec("curr[2].append(" + curr[0] + curr[1] + "_return(last_date,avg_price))")
			exec("print(\"The return for " + curr[0] + curr[1] + " is:" + str(curr[2][-1].hist_return) + " \")")

			if len(curr[2]) > 5:
				try:
					avg_pop_value = curr[2][-6].hist_return
				except:
					avg_pop_value = 0
				if isnan(avg_pop_value) == True:
					avg_pop_value = 0
			else:
				avg_pop_value = 0

			print(curr[2])

			# Calculate the average return value and print it/store it
			curr_avg = curr[2][-1].get_avg(avg_pop_value)
			exec("print(\"The average return for " + curr[0] + curr[1] + " is:" + str(curr_avg) + " \")")

			# Now that we have the average return, loop through the last 5 rows in the list to start compiling the
			# data needed to calculate the standard deviation
			for row in curr[2][-5:]:
				row.add_to_running_squared_sum(curr_avg)

			# Calculate the standard dev using the avg
			curr_std = curr[2][-1].get_std()
			exec("print(\"The standard deviation of the return for " + curr[0] + curr[1] + " is:" + str(
				curr_std) + " \")")

			# Calculate the average standard dev
			if len(curr[2]) > 5:
				try:
					pop_value = curr[2][-6].std_return
				except:
					pop_value = 0
			else:
				pop_value = 0
			curr_avg_std = curr[2][-1].get_avg_std(pop_value)
			exec("print(\"The average standard deviation of the return for " + curr[0] + curr[1] + " is:" + str(
				curr_avg_std) + " \")")

			print(curr[2])

			# -------------------Investment Strategy-----------------------------------------------
			try:
				return_value = curr[2][-1].hist_return
			except:
				return_value = 0
			if isnan(return_value) == True:
				return_value = 0

			try:
				return_value_1 = curr[2][-2].hist_return
			except:
				return_value_1 = 0
			if isnan(return_value_1) == True:
				return_value_1 = 0

			try:
				return_value_2 = curr[2][-3].hist_return
			except:
				return_value_2 = 0
			if isnan(return_value_2) == True:
				return_value_2 = 0

			try:
				upp_band = curr[2][-1].avg_return + (1.5 * curr[2][-1].std_return)
				if return_value >= upp_band and curr[
					3].Prev_Action_was_Buy == True and return_value != 0:  # (return_value > 0) and (return_value_1 > 0) and
					curr[3].sell_curr(avg_price)
			except:
				pass

			try:
				loww_band = curr[2][-1].avg_return - (1.5 * curr[2][-1].std_return)
				if return_value <= loww_band and curr[
					3].Prev_Action_was_Buy == False and return_value != 0:  # and  (return_value < 0) and (return_value_1 < 0)
					curr[3].buy_curr(avg_price)
			except:
				pass


# We can buy, sell, or do nothing each time we make a decision.
# This class defies a nobject for keeping track of our current investments/profits for each currency pair
class portfolio(object):
	def __init__(self, from_, to):
		# Initialize the 'From' currency amont to 1
		self.amount = 1
		self.curr2 = 0
		self.from_ = from_
		self.to = to
		# We want to keep track of state, to see what our next trade should be
		self.Prev_Action_was_Buy = False

	# This defines a function to buy the 'To' currency. It will always buy the max amount, in whole number
	# increments
	def buy_curr(self, price):
		if self.amount >= 1:
			num_to_buy = floor(self.amount)
			self.amount -= num_to_buy
			self.Prev_Action_was_Buy = True
			self.curr2 += num_to_buy * price
			print(
				"Bought %d worth of the target currency (%s). Our current profits and losses in the original currency (%s) are: %f." % (
					num_to_buy, self.to, self.from_, (self.amount - 1)))
		else:
			print("There was not enough of the original currency (%s) to make another buy." % self.from_)

	# This defines a function to sell the 'To' currency. It will always sell the max amount, in a whole number
	# increments
	def sell_curr(self, price):
		if self.curr2 >= 1:
			num_to_sell = floor(self.curr2)
			self.amount += num_to_sell * (1 / price)
			self.Prev_Action_was_Buy = False
			self.curr2 -= num_to_sell
			print(
				"Sold %d worth of the target currency (%s). Our current profits and losses in the original currency (%s) are: %f." % (
					num_to_sell, self.to, self.from_, (self.amount - 1)))
		else:
			print("There was not enough of the target currency (%s) to make another sell." % self.to)


def get_keltner_upper_bands(EMA, ATR, n):
	# Keltner Channel Upper Band = EMA + n*0.025*ATR, n from 1 to 100
	list = []
	for i in range(n):
		list.append(EMA + (i + 1) * 0.025 * ATR)
	return list


def get_keltner_lower_bands(EMA, ATR, n):
	# Keltner Channel Lower Band = EMA - n*0.025*ATR, n from 1 to 100
	list = []
	for i in range(n):
		list.append(EMA - (i + 1) * 0.025 * ATR)
	return list

def get_keltner_max_upper_bands(EMA, ATR, n):
	# Keltner Channel Upper Band = EMA + n*0.025*ATR, n from 1 to 100
	return EMA + n * 0.025 * ATR

def get_keltner_min_lower_bands(EMA, ATR, n):
	# Keltner Channel Upper Band = EMA + n*0.025*ATR, n from 1 to 100
	return EMA - n * 0.025 * ATR


def get_process(currency_pairs, db_path, total_run_time, aggregate_time):
	# The api key given by the professor
	key = "beBybSi8daPgsTp5yx5cHtHpYcrjp5Jq"  # from prof
	#     key = "IupyjcE9TmpZKrpvAdDFeuf1VT9whkFi"   # mine apikey

	# Number of list iterations - each one should last about 1 second
	count = 0
	agg_count = 0

	# Create an engine to connect to the database; setting echo to false should stop it from logging in std.out
	path = "sqlite+pysqlite:///" + db_path
	# engine = create_engine("sqlite+pysqlite:///sqlite_hw1/test.db", echo=False, future=True)
	engine = create_engine(path, echo=False, future=True)

	# Added
	#     engine = engine.connect()
	# Create the needed tables in the database

	initialize_raw_data_tables(engine, currency_pairs)  # CHANGED
	#     with engine.connect() as connection:
	#     with engine.begin() as conn:
	#         for curr in currency_pairs:
	#             print("executing create!!!!!!!!!!!!!!!!")
	#             conn.execute(text("CREATE TABLE "+curr[0]+curr[1]+"_raw(ticktime text, fxrate  numeric, inserttime text);"))

	initialize_aggregated_tables(engine, currency_pairs)

	# Open a RESTClient for making the api calls

	#     client = RESTClient(key)
	forex_client = polygon.ForexClient(key)  # for usual sync client
	print("connected to ForexClient")
	#     with RESTClient(key) as client:
	#         print("connected to restClient")
	# while count < 600:  # 86400 seconds = 24 hours  # 600 for test


	while count < total_run_time:  # 86400 seconds = 24 hours

		# Make a check to see if 6 minutes has been reached or not
		# if agg_count == 360:   # 20 for test
		if agg_count == aggregate_time:
			# Aggregate the data and clear the raw data tables

			aggregate_raw_data_tables(engine, currency_pairs)
			reset_raw_data_tables(engine, currency_pairs)
			agg_count = 0

		# Only call the api every 1 second, so wait here for 0.75 seconds, because the
		# code takes about .15 seconds to run
		time.sleep(0.75)

		# Increment the counters
		count += 1
		agg_count += 1

		# Loop through each currency pair
		for currency in currency_pairs:
			# Set the input variables to the API
			from_ = currency[0]
			to = currency[1]

			# Call the API with the required parameters
			try:
				#                 resp = client.forex_currencies_real_time_currency_conversion(from_, to, amount=100, precision=2)
				#                 resp = client.forex_currencies_real_time_currency_conversion(from_, to, amount=10, precision=2)
				resp = forex_client.real_time_currency_conversion(from_, to, amount=100, precision=2)
			except Exception as e:
				print(e)
				print("forex_client.real_time_currency_conversion() fail")
				continue

			#             print(type(resp))
			#             print(resp)
			# This gets the Last Trade object defined in the API Resource
			last_trade = resp['last']

			# Format the timestamp from the result
			dt = ts_to_datetime(last_trade["timestamp"])

			# Get the current time and format it
			insert_time = datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S')

			# Calculate the price by taking the average of the bid and ask prices
			avg_price = (last_trade['bid'] + last_trade['ask']) / 2

			#             conn = engine.begin()
			#             # Write the data to the SQLite database, raw data tables
			# #             print("INSERT INTO "+from_+to+"_raw(ticktime, fxrate, inserttime) VALUES ({0}, {1}, {2})").format( dt, avg_price, insert_time)
			#             conn.execute(text("INSERT INTO "+from_+to+"_raw(ticktime, fxrate, inserttime) VALUES (:ticktime, :fxrate, :inserttime)"),[{"ticktime": dt, "fxrate": avg_price, "inserttime": insert_time}])

			with engine.begin() as conn:
				#                 print("INSERT INTO "+from_+to+"_raw(ticktime, fxrate, inserttime) VALUES ({0}, {1}, {2})").format( dt, avg_price, insert_time)
				#                 conn.execute(text("INSERT INTO "+from_+to+"_raw(ticktime, fxrate, inserttime) VALUES (:ticktime, :fxrate, :inserttime)"),[{"ticktime": dt, "fxrate": avg_price, "inserttime": insert_time}])
				try:
					conn.execute(text(
						"INSERT INTO " + from_ + to + "_raw(ticktime, fxrate, inserttime) VALUES (:ticktime, :fxrate, :inserttime)"),
					             [{"ticktime": dt, "fxrate": avg_price, "inserttime": insert_time}])
				#                     print("inserting")
				except Exception as e:
					print(e)

	print("created")


# nullPeriod means in this period don't do anything, here is 20 (periods), which is
# so after 20 periods, we do the calculation
def get_process_agg(currency, db_path, nullPeriod, total_run_time, aggregate_time):
	# The api key given by the professor
	key = "beBybSi8daPgsTp5yx5cHtHpYcrjp5Jq"  # from prof
	#     key = "IupyjcE9TmpZKrpvAdDFeuf1VT9whkFi"   # mine apikey

	# Number of list iterations - each one should last about 1 second
	count = 0
	agg_count = 0

	# Create an engine to connect to the database; setting echo to false should stop it from logging in std.out
	path = "sqlite+pysqlite:///" + db_path
	# engine = create_engine("sqlite+pysqlite:///sqlite_hw1/test.db", echo=False, future=True)
	engine = create_engine(path, echo=False, future=True)

	# initialize the database
	initialize_aggfive_tables(engine, currency)

	# Open a RESTClient for making the api calls
	forex_client = polygon.ForexClient(key)  # for usual sync client
	print("connected to ForexClient")

	# lastMin = sys.float_info.max
	# lastMax = sys.float_info.min
	# EMA = 0
	# ATR = 0

	curMin = sys.float_info.max
	curMax = sys.float_info.min
	curMean = 0
	curTotal = 0
	curVOL = 0
	fd = 0
	lastPrice = 0

	while count < total_run_time:  # 86400 seconds = 24 hours
		from_ = currency[0]
		to = currency[1]

		# Increment the counters
		count += 1
		agg_count += 1

		# Call the API with the required parameters
		try:
			#                 resp = client.forex_currencies_real_time_currency_conversion(from_, to, amount=100, precision=2)
			#                 resp = client.forex_currencies_real_time_currency_conversion(from_, to, amount=10, precision=2)
			resp = forex_client.real_time_currency_conversion(from_, to, amount=100, precision=2)
		except Exception as e:
			print(e)
			print("forex_client.real_time_currency_conversion() fail")
			continue

		# Only call the api every 1 second, so wait here for 0.75 seconds, because the
		# code takes about .15 seconds to run
		time.sleep(0.75)

		#             print(type(resp))
		#             print(resp)
		# This gets the Last Trade object defined in the API Resource
		last_trade = resp['last']

		# Calculate the price by taking the average of the bid and ask prices
		avg_price = (last_trade['bid'] + last_trade['ask']) / 2

		# for each second update current min max, mean and vol
		curMin = min(curMin, avg_price)
		curMax = max(curMax, avg_price)
		curTotal += avg_price
		curMean = 1.0 * curTotal / agg_count
		curVOL = curMax - curMin
		print("curMin: \n", curMin)
		print("curMax: \n", curMax)
		# print("curTotal: \n", curTotal)
		# print("curMean: \n", curMean)
		print("curVOL: \n", curVOL)


		# if can calculate keltner bands, calculate fd for each 6-minute bucket
		# after inserting, means the new start of 6 minutes
		if count > nullPeriod:
			# each interval between bands are 0.025*ATR
			# to know how many bands it crossed is to count the difference between curPrice and lastPrce is composed of how many intervals
			interval = 0.025 * ATR
			if interval != 0:
				larger = max(avg_price, lastPrice)
				smaller = min(avg_price, lastPrice)
				maxLine = min(get_keltner_max_upper_bands(EMA, ATR, 100), larger)
				minLine = max(get_keltner_min_lower_bands(EMA, ATR, 100), smaller)
				print("maxLine: ", maxLine)
				print("minLine: ", minLine)
				print("interval: \n", interval)
				if int((maxLine - minLine) / interval) < 0:
					print("smallerrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrr")
				fd += int((maxLine - minLine) / interval)

		lastPrice = avg_price

		if agg_count == aggregate_time:
			# when it reaches the end of the bucket, update and reinitialize information for each bucket
			if count < nullPeriod:
				agg_count = 0
			else:
				# Aggregate the data and clear the raw data tables
				lastMin = curMin
				lastMax = curMax
				EMA = curMean  # lastMean
				ATR = curVOL
				print("ATR: ", ATR)
				keltner_upper_bands = get_keltner_upper_bands(EMA, ATR, 100)
				keltne_lower_bands = get_keltner_lower_bands(EMA, ATR, 100)
				# when count == nullPeriod, it means the end of the nullPeriod,
				# only calculate EMA, ATR and bands for calculating fd for the next period

				if count > nullPeriod:
					# update and insert
					lastfd = fd
					# Get the current time and format it
					insert_time = datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S')
					# add column (max, min, mean, vol, fd, inserttime) for each aggregate_time minutes
					with engine.begin() as conn:
						try:
							print("inserting: max {0}, min {1}, mean {2}, vol {3}, fd {4}, inserttime {5}\n".format(lastMax, lastMin, EMA, ATR, lastfd, insert_time))
							conn.execute(text(
								"INSERT INTO " + from_ + to + "_aggfive(max, min, mean, vol, fd, inserttime) VALUES (:max, :min, :mean, :vol, :fd, :inserttime)"),
								[{"max": lastMax, "min": lastMin, "mean": EMA, "vol": ATR, "fd": lastfd,
								  "inserttime": insert_time}])
						#                     print("inserting")
						except Exception as e:
							print(e)

			# since we calculate for each bucket, we need to reinitialize
			curMin = sys.float_info.max
			curMax = sys.float_info.min
			curMean = 0
			curTotal = 0
			agg_count = 0
			fd = 0



	print("created")
